<?php
#This code was written by Tom Payerle payerle@physics.umd.edu
#(c) 2006 University of Maryland
#Lower level CAS stuff based on phpCAS code by ESUP-portail consortium
#http://www.esup-portail.org, which is under the LGPL
#
#This code is free software, you can redistribute it and/or modify if under
#the terms of the Lesser General Public Library (LGPL) as published by
#the Free Software Foundation.  See http://www.gnu.org/licenses/lgpl.html
#for details.  A copy should also be present with this code (see the file
#COPYRIGHT.txt)
#
#SimpleCAS Version 1.0 Sept 26, 2006
#$Id:$

#Constant definitions

#CAS versions
define("SCAS_CAS_1_0",'1.0');
define("SCAS_CAS_2_0",'2.0');
                                                                                
#Default debug level
define("SCAS_DEF_DEBUG_LEVEL",0);

#Debug levels for certain types of info
define("SCAS_DBG_AUTHENTICATE1", 100);
define("SCAS_DBG_AUTHENTICATE2", 200);

define("SCAS_DBG_REDIRECT1", 300);
define("SCAS_DBG_AUTHINFO1", 400);
define("SCAS_DBG_TICKET1", 500);
define("SCAS_DBG_REDIRECT2", 600);
define("SCAS_DBG_AUTHINFO2", 700);
define("SCAS_DBG_TICKET2", 800);
	
define("SCAS_DBG_INIT1", 1000);
define("SCAS_DBG_INIT2", 1100);

class SimpleCAS 
{	#Data members, representing various option settings

	#CAS Server options
	var $casVersion; #Version of CAS to use
	var $casServer; #IP address/hostname of CAS server
	var $casPort;	#Port CAS server is listening on
	var $casPath;	#path portion of base URL to CAS
	var $casCAInfo; #Path to CA cert for CAS server
	var $casCAPath; #Path to directory containing CA certs (for CAS server)
	var $casVerifyPeer; #Whether we must verify cert of CAS server
	var $casVerifyHost; #What level to verify hostname in CAS cert,
			#see curl option CURLOPT_SSL_VERIFYHOST

	#Constructor behavior options
	var $autoStartSession; #Whether to start session automatically
	var $doNotAutoAuthenticate; #Whether "new" should automatically 
				#authenticate

	#Authentication/logout method options
	var $authenticationOptional; #Whether to use CAS gateway mode
	var $authOptDeltaTime; #Time (in seconds) before visiting CAS again
		#in authenticate_optional if no valid credentials
	var $sessionVarNameOptTstamp; #name of sess var for tracking last
		#visit to CAS server in optional mode
	var $forcePassword; #Whether to use CAS renew mode
	var $casLogoutOnLogout; #Whether logout deletes CAS TGT cookie
	var $destroySessionOnLogout; #Whether logout destroys whole session

	#Session authentication information options
	var $sessionName; #Name of the session/cookie containing session id
	var $sessionVarName; #Name of session variable for authInfo
	var $authInfoExpiry; #How long (in seconds) auth from session valid
	var $authInfoExpiryLastUse; #How long (in seconds) auth from session 
			#valid since last used
	var $forceExpiry; #Absolute time that force flag in session valid
	var $forceExpiryLastUse; #Max Time that force flag in session valid
			#since last used (in authenticate_forced)
	var $authInfoSameIP; #Whether auth from session must come from same IP
	var $autoChangeSessionIDs; #If true, change session id every time
		#we generate an authinfo hash in the session.

	#This one gets a default in case want to use before object fully
	#initialized
	var $debugLevel = SCAS_DEF_DEBUG_LEVEL;


#--------------------------------------------------------------------
#		Logging, error reporting, etc. methods
#--------------------------------------------------------------------

	function _log_filter($message)
	#Filters log messages
	#Adds '> ' after any newlines to avoid fake log messages
	{	$message = ereg_replace("\n","\n> ", $message);
		return $message;
	}
		
	function _log_message($message)
	#Logs a message (prints to stderr with appropriate prefixes)
	{	$now = time();
		$nowstr = date("Y-m-d H:i:s", $now);
		$app='SimpleCAS';
		$rmhost = $_SERVER['REMOTE_ADDR'];
		$header="$nowstr $app [$rmhost] ";
		$fd = fopen('php://stderr','w');
		$message = $this->_log_filter($message);
		fwrite($fd, $header . $message . "\n");
	}

	function debug($message,$level=0)
	#logs a debugging message if $level < $this->debugLevel
	{	if ( $level <= $this->debugLevel)
		{	$this->_log_message("debug: $message");
		}
	}

	function _fatal_error_html($message)
	#Produces the error page for the user for a fatal error
	#$message is the message that goes into the logs, it is
	#ignored in the standard version but provided in case desired
	#in an overridden version
	{
?>
<html>
<header><title>Fatal Error</title>
</header>
<body>
<h1>Fatal Error</h1>
<p>A fatal error occurred in your application.  Please contact PCS staff
to inform them of this problem.  Thank you.

</body>
<?php
	}

	function fatal_error($message)
	#Logs a fatal error, then prints an apology to user and aborts.
	{	$this->_log_message("FATAL ERROR: $message");
		$this->_fatal_error_html(htmlentities($message));
		exit();
	} 

	function error($message)
	#Logs an error message
	{	$this->_log_message("ERROR: $message");
	}

	function warning($message)
	#Logs a warning message
	{	$this->_log_message("Warning: $message");
	}

#--------------------------------------------------------------------
#		Constructors, initializers, etc
#--------------------------------------------------------------------

	function _initObject($args)
	#Initialize a newly created argument with named based arguments
	#Versions overridden in subclasses should probably remove any args
	#they process and then call this version
	{	$this->debug("Starting SimpleCAS initObject", SCAS_DBG_INIT1);
		foreach ( $args as $key => $value )
		{	$key = strtolower($key);
			$this->debug("About to set $key => $value", 
				SCAS_DBG_INIT2);
			switch ($key)
			{ case 'casversion':
				$this->casVersion=$value; break;
			  case 'casserver':
				$this->casServer = $value; break;
			  case 'caspath':
				$this->casPath = $value; break;
			  case 'autostartsession':
				$this->autoStartSession = $value; break;
			  case 'donotautoauthenticate':
				$this->doNotAutoAuthenticate = $value; break;
			  case 'authenticationoptional':
				$this->authenticationOptional= $value; break;
			  case 'forcepassword':
				$this->forcePassword = $value; break;
			  case 'forceexpiry':
				$this->forceExpiry = $value; break;
			  case 'forceexpirylastuse':
				$this->forceExpiryLastUse = $value; break;
			  case 'caslogoutonlogout':
				$this->casLogoutOnLogout = $value; break;
			  case 'destroysessiononlogout':
				$this->destroySessionOnLogout = $value; break;
			  case 'autochangesessionids':
				$this->autoChangeSessionIDs= $value; break;
			  case 'authinfoexpiry':
				$this->authInfoExpiry= $value; break;
			  case 'authinfosameip':
				$this->authInfoSameIP = $value; break;
			  case 'cascainfo':
				$this->casCAInfo = $value; break;
			  case 'cascapath':
				$this->casCAPath = $value; break;
			  case 'casverifypeer':
				$this->casVerifyPeer = $value; break;
			  case 'casverifyhost':
				$this->casVerifyHost = $value; break;
			  case 'authoptdeltatime':
				$this->authOptDeltaTime = $value; break;
			  case 'sessionname':
				$this->sessionName = $value; break;
			  case 'sessionvarname':
				$this->sessionVarName = $value; break;
			  case 'sessionvarnameopttstamp':
				$this->sessionVarNameOptTstamp = $value; break;
			  case 'debuglevel':
				$this->debugLevel = $value; break;
			  default:
				$this->fatal_error("invalid arg key $key");
				die("Should not reach here");
			}
		}
		$this->debug("Leaving SimpleCAS initObject",SCAS_DBG_INIT1);
	}

	function _defaultSettings()
	#Supply default values to any settings not set via arguments
	#	casVersion defaults to SCAS_CAS_2_0
	#	casPort defaults to 443 (standard https: port)
	#	casPath: default to ''
	#	autoStartSession defaults to TRUE
	#	doNotAutoAuthenticate defaults to FALSE
	#	forcePassword defaults to FALSE
	#	forceExpiry defaults to 1800 seconds (30 minutes)
	#	forceExpiryLastUse defaults to 1800 seconds (30 minutes)
	#	authenticationOptional defaults to FALSE
	#	casLogoutOnLogout defaults to FALSE
	#	destroySessionOnLogout defaults to FALSE
	#	autoChangeSessionIDs defaults to TRUE
	#	authInfoExpiry defaults to 1800 seconds (30 minutes)
	#	authInfoExpiryLastUse defaults to 1800 seconds (30 minutes)
	#	authInfoSameIP defaults to TRUE
	#	casVerifyPeer defaults to TRUE
	#	casVerifyHost defaults to 2
	#	authOptDeltaTime defaults to 300 (seconds = 5 minutes)
	#	sessionName = "PHPSESSID"
	#	sessionVarName="__authinfo";
	#	sessionVarNameOptTstamp="__authinfo_optTstamp";
	#Versions overridden in subclasses should probably default anything
	#they want to default then call this version
	{	$this->debug("Starting SimpleCas defaultSettings", 
			SCAS_DBG_INIT1);
		if (! IsSet($this->casVersion) )
		{	$this->casVersion = SCAS_CAS_2_0;
			$this->debug("default casVersion to " 
				. $this->casVersion, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->casPort) )
		{	$this->casPort = 443;
			$this->debug("default casPort to " 
				. $this->casPort, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->casPath) )
		{	$this->casPath = '';
			$this->debug("default casPath to " 
				. $this->casPath, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->autoStartSession) )
		{	$this->autoStartSession = TRUE;
			$this->debug("default autoStartSession to " 
				. $this->autoStartSession, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->doNotAutoAuthenticate) )
		{	$this->doNotAutoAuthenticate = FALSE;
			$this->debug("default doNotAutoAuthenticate to " 
				. $this->doNotAutoAuthenticate, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->forcePassword) )
		{	$this->forcePassword = FALSE;
			$this->debug("default forcePassword to " 
				. $this->forcePassword, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->forceExpiry) )
		{	$this->forceExpiry = 1800; #1800 seconds= 30 min
			$this->debug("default forceExpiry to " 
				. $this->forceExpiry, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->forceExpiryLastUse) )
		{	$this->forceExpiryLastUse = 1800; #1800 secs= 30 min
			$this->debug("default forceExpiryLastUse to " 
				. $this->forceExpiryLastUse, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->authenticationOptional) )
		{	$this->authenticationOptional = FALSE;
			$this->debug("default authenticationOptional to " .
				$this->authenticationOptional, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->casLogoutOnLogout) )
		{	$this->casLogoutOnLogout = FALSE;
			$this->debug("default casLogoutOnLogout to " 
				. $this->casLogoutOnLogout, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->destroySessionOnLogout) )
		{	$this->destroySessionOnLogout = FALSE;
			$this->debug("default destroySessionOnLogout to " 
			    . $this->destroySessionOnLogout, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->autoChangeSessionIDs) )
		{	$this->autoChangeSessionIDs = TRUE;
			$this->debug("default autoChangeSessionIDs to " 
				. $this->autoChangeSessionIDs, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->authInfoExpiry) )
		{	$this->authInfoExpiry = 1800; #1800 seconds = 30 min
			$this->debug("default authInfoExpiry to " 
				. $this->authInfoExpiry, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->authInfoExpiryLastUse) )
		{	$this->authInfoExpiryLastUse = 1800; #1800 secs= 30 min
			$this->debug("default authInfoExpiryLastUse to " .
				$this->authInfoExpiryLastUse, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->authInfoSameIP) )
		{	$this->authInfoSameIP = TRUE;
			$this->debug("default authInfoSameIP to " 
				. $this->authInfoSameIP, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->casVerifyPeer) )
		{	$this->casVerifyPeer = TRUE;
			$this->debug("default casVerifyPeer to " 
				. $this->casVerifyPeer, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->casVerifyHost) )
		{	$this->casVerifyHost = TRUE;
			$this->debug("default casVerifyHost to " 
				. $this->casVerifyHost, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->authOptDeltaTime) )
		{	$this->authOptDeltaTime = 300;#300 seconds = 5 min
			$this->debug("default authOptDeltaTime to " 
				. $this->authOptDeltaTime, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->sessionName) )
		{	$this->sessionName="PHPSESSID";
			$this->debug("default sessionName to " 
				. $this->sessionName, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->sessionVarName) )
		{	$this->sessionVarName="__authinfo";
			$this->debug("default sessionVarName to " 
				. $this->sessionVarName, SCAS_DBG_INIT2);
		}
		if (! IsSet($this->sessionVarNameOptTstamp) )
		{	$this->sessionVarNameOptTstamp="__authinfo_optTstamp";
			$this->debug("default sessionVarNameOptTstamp to " .
			    $this->sessionVarNameOptTstamp, SCAS_DBG_INIT2);
		}
		$this->debug("Leaving SimpleCas defaultSettings",
			SCAS_DBG_INIT1);
	}

	function _checkRequiredArgs()
	#Generate a fatal_error if any required arguments are still missing
	#We require the following to be set:
	#	casVersion
	#	autoStartSession
	#	casServer
	#	casPort
	#	casPath
	#	doNotAutoAuthenticate
	#	forcePassword
	#	forceExpiry
	#	forceExpiryLastUse
	#	authenticationOptional
	#	casLogoutOnLogout
	#	destroySessionOnLogout
	#	autoChangeSessionIDs
	#	authInfoExpiry
	#	authInfoExpiryLastUse
	#	authInfoSameIP
	#	casVerifyPeer
	#	casVerifyHost
	#	authOptDeltaTime
	#	sessionName
	#	sessionVarName
	#	sessionVarNameOptTstamp
	#We also do some sanity checking on values for
	#	casVersion
	#	casServer, casPort, authInfoExpiry, casVerifyHost
	#	authInfoExpiryLastUse, forceExpiry, forceExpiryLastUse,
	#	sessionVarName, sessionVarNameOptTstamp, sessionName
	#	(most of rest are just eval'ed as boolean, or casPath is
	#	just part of URL, and optionally null)
	#Versions overridden in subclasses should probably check anything
	#they are concerned about not on this list and then call us.
	{	$this->debug("Starting SimpleCas checkrequiredArgs",
			SCAS_DBG_INIT1);
		if ( ! IsSet($this->casVersion) )
		{	$this->fatal_error("casVersion is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->autoStartSession) )
		{	$this->fatal_error("autoStartSession is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->casServer) )
		{	$this->fatal_error("casServer is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->casPort) )
		{	$this->fatal_error("casPort is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->casPath) )
		{	$this->fatal_error("casPath is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->doNotAutoAuthenticate) )
		{	$this->fatal_error("doNotAutoAuthenticate is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->forcePassword) )
		{	$this->fatal_error("forcePassword is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->forceExpiry) )
		{	$this->fatal_error("forceExpiry is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->forceExpiryLastUse) )
		{	$this->fatal_error("forceExpiryLastUse is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->authenticationOptional) )
		{	$this->fatal_error("authenticationOptional is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->casLogoutOnLogout) )
		{	$this->fatal_error("casLogoutOnLogout is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->destroySessionOnLogout) )
		{	$this->fatal_error("destroySessionOnLogout is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->autoChangeSessionIDs) )
		{	$this->fatal_error("autoChangeSessionIDs is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->authInfoExpiry) )
		{	$this->fatal_error("authInfoExpiry is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->authInfoExpiryLastUse) )
		{	$this->fatal_error("authInfoExpiryLastUse is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->authInfoSameIP) )
		{	$this->fatal_error("authInfoSameIP is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->casVerifyPeer) )
		{	$this->fatal_error("casVerifyPeer is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->casVerifyHost) )
		{	$this->fatal_error("casVerifyHost is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->authOptDeltaTime) )
		{	$this->fatal_error("authOptDeltaTime is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->sessionName) )
		{	$this->fatal_error("sessionName is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->sessionVarName) )
		{	$this->fatal_error("sessionVarName is not set");
			die("Should not reach here");
		}
		if ( ! IsSet($this->sessionVarNameOptTstamp) )
		{	$this->fatal_error("sessionVarNameOptTstamp is not set");
			die("Should not reach here");
		}

		#Validate casServer, can be numeric IP address or hostname
		if ( ereg('^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)$',
			$this->casServer, $match) )
		{	#Looks like an numeric IP address
			#Check all numbers between 0 and 255
			for($i=0; $i<4; $i++)
			{	if ( $match[$i] <0 || $match[$i]> 255 )
				{  $this->fatal_error( "invalid octet in IP " .
					"address for casServer: " . 
					$this->casServer);
				    die("Should not reach here");
				}
			}
		} else
		{	#Looks like hostname was given
			#Make sure at least 3 parts (2 dots), and all
			#parts consist only of letters, digits, and -
			$match = explode('.', $this->casServer);
			if ( count($match) < 3 )
			{	$this->fatal_error("Too few elements in " .
					"hostname for casServer: " . 
					$this->casServer);
				die("Should not reach here");
			}
			for($i=0; $i< count($match); $i++)
			{	if (! ereg('^[a-zA-Z0-9\-]+$', $match[$i]) )
				{	$this->fatal_error("Illegal chars in hostname for casServer: " . $this->casServer);
					die("Should not reach here");
				}
			}
		}

		#Validate casVersion is SCAS_CAS_1_0 or 2_0
		if ( ! ($this->casVersion == SCAS_CAS_1_0 ||
			$this->casVersion == SCAS_CAS_2_0) )
		{	$this->fatal_error("Invalid casVersion " . 
				$this->casVersion);
			die("Should not reach here");
		}

		#Ensure a valid port number in casPort
		if ( ! (is_int($this->casPort) && $this->casPort > 0 &&
			$this->casPort < 65536 ) )
		{	$this->fatal_error("Invalid casPort: " . 
				$this->casPort);
			die("Should not reach here");
		}

		#Ensure a valid authInfoExpiry
		if ( ! (is_int($this->authInfoExpiry) && 
				$this->authInfoExpiry > -1 ) )
		{	$this->fatal_error("Invalid authInfoExpiry: " . 
			$this->authInfoExpiry);
			die("Should not reach here");
		}

		#Ensure a valid authInfoExpiryLastUse
		if ( ! (is_int($this->authInfoExpiryLastUse) && 
				$this->authInfoExpiryLastUse > -1 ) )
		{	$this->fatal_error("Invalid authInfoExpiryLastUse: " . 
				$this->authInfoExpiry);
			die("Should not reach here");
		}

		#Ensure a valid VerifyHost value
		if ( ! (is_int($this->casVerifyHost) && 
				$this->casVerifyHost >= 0 &&
				$this->casVerifyHost < 3 ) )
		{	$this->fatal_error("Invalid casVerifyHost: " . 
				$this->casVerifyHost);
			die("Should not reach here");
		}

		#Ensure a valid authOptDeltaTime value
		if ( ! (is_int($this->authOptDeltaTime) && 
				$this->authOptDeltaTime > 0 ) )
		{	$this->fatal_error("Invalid authOptDeltaTime: " . 
				$this->authOptDeltaTime);
			die("Should not reach here");
		}

		#Ensure a valid forceExpiry
		if ( ! (is_int($this->forceExpiry) && 
				$this->forceExpiry > -1 ) )
		{	$this->fatal_error("Invalid forceExpiry: " . 
			$this->forceExpiry);
			die("Should not reach here");
		}

		#Ensure a valid forceExpiryLastUse
		if ( ! (is_int($this->forceExpiryLastUse) && 
				$this->forceExpiryLastUse > -1 ) )
		{	$this->fatal_error("Invalid forceExpiryLastUse: " . 
				$this->forceExpiry);
			die("Should not reach here");
		}

		#Ensure a valid sessionName, must be composed only of
		#digits and letters
		if ( ! ereg('^[a-zA-Z0-9]+$',$this->sessionName) )
		{	$this->fatal_error("Invalid sessionName: " . 
			$this->sessionName);
			die("Should not reach here");
		}

		#Ensure a valid sessionVarName, must be composed only of
		#digits, letters, underscores
		if ( ! ereg('^[a-zA-Z0-9\_]+$',$this->sessionVarName) )
		{	$this->fatal_error("Invalid sessionVarName: " . 
			$this->sessionVarName);
			die("Should not reach here");
		}

		#Same for sessionVarNameOptTstamp
		if ( ! ereg('^[a-zA-Z0-9\_]+$',$this->sessionVarNameOptTstamp))
		{	$this->fatal_error("Invalid sessionVarNameOptTstamp: " 
			.  $this->sessionVarNameOptTstamp);
			die("Should not reach here");
		}
		$this->debug("Leaving SimpleCas checkrequiredArgs",
			SCAS_DBG_INIT1);
	}

	function _session_init()
	#Initialize session if requested
	#Mainly provided so you can overload, we just do a basic session_start
	#
	#This is also invoked by _change_session_id to start up the new session
	#after we destroy the old one.
	{	session_name($this->sessionName);
		session_start();
	}

	function _constructObject($args=array())
	#Our basic constructor, with wrong name
	{	$this->debug("Starting SimpleCas constructObject",
			SCAS_DBG_INIT1);
		$this->_initObject($args);
		$this->_defaultSettings();
		$this->_checkRequiredArgs();

		if ( $this->autoStartSession )
		{	$this->_session_init();
		}
		if ( ! $this->doNotAutoAuthenticate )
		{	$this->debug("constructObject about to authenticate",
				SCAS_DBG_INIT2);
			$this->authenticate();
		}
		$this->debug("Leaving SimpleCas constructObject",
			SCAS_DBG_INIT1);
	}

	function SimpleCAS($args=array())
	#PHP4 constructor
	{	$this->_constructObject($args);
	}

	function __construct($args=array())
	#PHP5 constructor
	{	$this->_constructObject($args);
	}

#--------------------------------------------------------------------
#		AuthInfo methods
#		We store some login information in session
#--------------------------------------------------------------------

	function _ip_compare($ip1,$ip2)
	#Compares 2 ip addresses, return TRUE if equal, false otherwise
	{	$ip1arr = explode('.', $ip1);
		$ip2arr = explode('.', $ip2);

		if ( $ip1arr[0] != $ip2arr[0] )
			return FALSE;
		if ( $ip1arr[1] != $ip2arr[1] )
			return FALSE;
		if ( $ip1arr[2] != $ip2arr[2] )
			return FALSE;
		if ( $ip1arr[3] != $ip2arr[3] )
			return FALSE;
		return TRUE;
	}

	function is_authinfo_valid($forced=FALSE)
	#Checks for valid authinfo from a previous session
	#Returns true (the username logged in as) if it is valid
	#Returns false otherwise, and may clear invalid/expired authinfo from
	#session
	#If $forced is true, the forced flag must also be set in the authinfo
	{	$this->debug("Starting is_authinfo_valid, forced=$forced",
			SCAS_DBG_AUTHINFO1);
		if ( ! IsSet($_SESSION[$this->sessionVarName]) )
		{	$this->debug("No authinfo, leaving is_authinfo_valid",
				SCAS_DBG_AUTHINFO1);
			return FALSE;
		}

		$sessionHash = $_SESSION[$this->sessionVarName];

		if ( IsSet($sessionHash['username']) )
		{	$username = $sessionHash['username'];
		} else
		{	$this->debug("No username, leaving is_authinfo_valid",
				SCAS_DBG_AUTHINFO1);
			$this->_clearAuthInfo();
			return FALSE;
		}

		$now = time();
		$ctime = $sessionHash['ctime'];
		if ( $this->authInfoExpiry != 0 &&
			$this->authInfoExpiry + $ctime < $now )
		{	$this->debug("Authinfo expired, leaving is_ai_valid",
				SCAS_DBG_AUTHINFO1);
			$this->_clearAuthInfo();
			return FALSE;
		}

		$ctime = $sessionHash['lastused'];
		if ( $this->authInfoExpiryLastUse != 0 &&
			$this->authInfoExpiryLastUse + $ctime < $now )
		{	$this->debug("Authinfo expired (Last Use), " .
				"leaving is_ai_valid",
				SCAS_DBG_AUTHINFO1);
			$this->_clearAuthInfo();
			return FALSE;
		}

		$ipfrom = $sessionHash['ipfrom'];
		$myIP = $_SERVER['REMOTE_ADDR'];
		if ( $this->authInfoSameIP 
			&& ! $this->_ip_compare($ipfrom, $myIP) )
		{	$this->warning("IP mismatch wanted $ipfrom, got $myIP,"
				. " leaving is_ai_valid");
			$this->_clearAuthInfo();
			return FALSE;
		}

		if ( $forced )
		{	if ( ! (IsSet($sessionHash['forced']) &&
				$sessionHash['forced'] ) )
			{	$this->debug("Not forced, but require it, so "
					. "leaving is_authinfo_valid",
					SCAS_DBG_AUTHINFO1);
				#$this->_clearAuthInfo();
				#Leave with valid unforced authinfo
				return FALSE;
			}

			$ctime = $sessionHash['ctime'];
			if ( $this->forceExpiry != 0 &&
				$this->forceExpiry + $ctime < $now )
			{	$this->debug("Authinfo force expired,"
					 .  "leaving is_ai_valid",
					SCAS_DBG_AUTHINFO1);
				$this->_unsetAuthInfoForced();
				return FALSE;
			}


			$ctime = $sessionHash['flastused'];
			if ( ! IsSet($ctime) )
			{	$this->warning("flastused unset, using 0");
				$ctime = 0;
			}
			if ( $this->forceExpiryLastUse != 0 &&
				$this->forceExpiryLastUse + $ctime < $now )
			{	$this->debug("Authinfo force expired,(lastUse)"
					 .  "leaving is_ai_valid",
					SCAS_DBG_AUTHINFO1);
				$this->_unsetAuthInfoForced();
				return FALSE;
			}
		}
			

		#Valid, return username
		$this->debug("is_authinfo_valid returning true ($username)",
			SCAS_DBG_AUTHINFO1);
		return $username;
	}

	function _change_session_id()
	#This uses session_regenerate_id to change session ids.  Invoked 
	#everytime we generateAuthInfo.
	#We take care to delete the old session as we are mainly concerned
	#with reducing risk of session_id hijacking
	#
	#Thanks to Nicolas Chachereau's posting of 2 June 2005 on php.net
	#man page for regenerate_session_id
	{	$old_id = session_id();
		session_regenerate_id();
		$new_id = session_id();
		$this->debug("Session id changed from $old_id to $new_id",
			SCAS_DBG_AUTHINFO2);
		session_id($old_id);
		unset($_SESSION[$this->sessionVarName]); #Overkill
		session_destroy();
		$tmp = $_SESSION;
		session_id($new_id);
		$this->_session_init();
		$_SESSION = $tmp;
	}

	function _generateAuthInfo($user,$forced=FALSE)
	#Generates the authinfo stuff for username $user
	#If $forced is set, mark the forced flag
	{	$this->debug("Called generateAuthInfo for user $user, mode=" .
			$forced, SCAS_DBG_AUTHINFO1);
		$myIP = $_SERVER['REMOTE_ADDR'];
		$now = time();
		
		if ( $this->autoChangeSessionIDs )
		{	$this->debug("About to change session ids",
				SCAS_DBG_AUTHINFO2);
			$this->_change_session_id();
		}
		$_SESSION[$this->sessionVarName] = 
			array( 'ipfrom' => $myIP,
				'ctime' => $now,
				'lastused' => $now,
				'username' => $user,
				'forced' => $forced,
				);
		if ( $forced )
		{ 	$_SESSION[$this->sessionVarName]['flastused']=$now;
		}
	}

	function _clearAuthInfo()
	#Clears the authinfo stuff
	{	$this->debug("Called clearAuthInfo", SCAS_DBG_AUTHINFO1);
		if ( ! IsSet($_SESSION[$this->sessionVarName]) )
			return;
		unset($_SESSION[$this->sessionVarName]);
	}

	function _updateAuthInfoLastUsed($forced=FALSE)
	#Updates the lastused timestamps on authInfo
	#If $forced is set, updates flastused as well
	{	$this->debug("Called updateAuthInfoLastUsed forced=$forced", 
			SCAS_DBG_AUTHINFO1);
		if ( ! IsSet($_SESSION[$this->sessionVarName]) )
		{	$this->warning("Attempt to update lastused on " .
				"null authInfo");
			return;
		}
		$now = time();
		$_SESSION[$this->sessionVarName]['lastused']=$now;
		if ( $forced )
		{ 	$_SESSION[$this->sessionVarName]['flastused']=$now;
		}
	}

	function _unsetAuthInfoForced()
	#Unset the "forced" flags in AuthInfo
	{	$this->debug("Called unsetAuthInfoForced",SCAS_DBG_AUTHINFO1);
		if ( ! IsSet($_SESSION[$this->sessionVarName]) )
		{	$this->warning("Attempt to unset forced on " .
				"null authInfo");
			return;
		}
		if ( ! $_SESSION[$this->sessionVarName]['forced'] )
		{	$this->warning("Attempt to unset forced on " .
				"non-forced authInfo");
		}
		$_SESSION[$this->sessionVarName]['forced']=FALSE;
		unset($_SESSION[$this->sessionVarName]['flastused']);
	}

#--------------------------------------------------------------------
#		Methods for lower level URL, HTTP handling
#--------------------------------------------------------------------

	function _str_split($text)
	#Mimics php5 str_split for split_lenght=1
	{ 	$arry=array();
		for($i=0; $i<strlen($text); $i++) 
		{	$arry[$i]=$text{$i};
		}
		return $arry;
	}

	function _URIEncodeString($string)
	#Takes a non URI encoded string and encodes it
	#	letters (upper or lowercase), digits, and -_.~ passed unchanged
	#	spaces become +
	#	any other ascii character becomes %HH where HH is the hex
	#		value of its ASCII code
	{	
		//$arry = str_split($string); #This is PHP5 only
		$arry = $this->_str_split($string);
		foreach ( $arry as $ind => $char )
		{	$asc = ord($char);
			#digits pass through unchanged
			if ( $asc >= ord('0') && $asc <= ord('9') )
				continue;
			#uppercase pass through unchanged
			if ( $asc >= ord('A') && $asc <= ord('Z') )
				continue;
			#lowercase pass through unchanged
			if ( $asc >= ord('a') && $asc <= ord('z') )
				continue;
			#chars '-_.~' pass through unchanged
			if ( $char == '-' || $char == '_' || $char == '.'
				|| $char == '~' )
				continue;

			#space becomes +
			if ( $char == ' ' )
			{	$arry[$ind] = '+';
				continue;
			}

			#Anything else in %HH notation
			$hex = DecHex($asc);
			if ( strlen($hex) < 2 )
			{	$hex = '0' . $hex;
			}
			$arry[$ind] = "%$hex";
		}
		$newstr = implode('',$arry);
		return $newstr;
	}

	function _constructURLWithArgs($baseURL,$parms=FALSE)
	#Generates an url from $baseURL with optional parameters attached
	#in GET fashion.
	# $baseURL will default to $this->myURL() if false
	# $parms should be a list of parameter names => values.  No parameters
	#	added if empty
	# Returns the URL as a string
	{	if ( ! $baseURL ) $baseURL = $this->myURL();
		if ( ! $parms ) $parms = array();
		$newurl = $baseURL;
		$delim = '?';
		foreach ( $parms as $key => $val )
		{	$key = $this->_URIEncodeString($key);
			$val = $this->_URIEncodeString($val);
			$newurl = "$newurl$delim$key=$val";
			$delim='&';
		}
		return $newurl;
	}

	function _curl_init($url,$cookies)
	#Function to initialize our cURL session to the url $url
	#$cookies is an array of cookies to be sent with the request
	#(the $cookies argument to _readURL)
	#Returns the initialized curl object
	#
	#_readURL expects the options CURLOPT_RETURNTRANSFER and
	#CURLOPT_HEADER to be set, so ensure you set them if you override
	#this function
	#
	#The SimpleCAS options casCAInfo, casCAPath, casVerifyPeer, and
	#casVerifyHost are also handled here, and if you override this you
	#should either handle yourself or call this version, or risk breaking
	#these options.
	#
	#The options casCAInfo and casCAPath set the location of certs for
	#the CA signing the CAS server cert
	#
	#Server(Peer) certs will be verified against CA if $casVerifyPeer is
	#TRUE.  
	#$casVerifyHost will verify hostname in cert at various levels (see
	#CURLOPT_VERIFYHOST).  May need to lower to 0 or 1 if not using
	#casVerifyPeer.
	{	
		#initialize the CURL session
		$curl = curl_init($url);

		#Set Curl options

		#verify the the server's certificate corresponds to its name
		#if verifyHost==2, just that have a CN in cert if 1
		#and nothing if 0.  May need to downgrade if not verifying peer
		curl_setopt($curl,CURLOPT_SSL_VERIFYHOST,
			$this->casVerifyHost);

		#verify the CAS server's certificate using CA if set
		curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, 
			$this->casVerifyPeer);
		#Some info abount CA certs
		if ( $this->casCAInfo )
		{	curl_setopt($curl, CURLOPT_CAINFO, $this->casCAInfo);
		}
		if ( $this->casCAPath )
		{	curl_setopt($curl, CURLOPT_CAPATH, $this->casCAPath);
		}

		#return the CURL output into a variable
		curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
		#include the HTTP header with the body
		curl_setopt($curl, CURLOPT_HEADER, 1);

		#add cookies headers
		if ( is_array($cookies) ) 
		{ 	curl_setopt($curl,CURLOPT_COOKIE,implode(';',$cookies));
		}

		return $curl;
	}

	function _readURL($url,$cookies,&$headers,&$body,&$err_msg,
		$verifyPeer=TRUE,$verifyHostname=2)
	#Method to access a remote URL
	#@param $url the URL to access.
	#@param $cookies an array containing cookies strings such as 'name=val'
	#@param $headers an array containing the HTTP header lines of the 
	#	response (an empty array on failure).
	#@param $body the body of the response, as a string (empty on failure).
	#@param $err_msg an error message, filled on failure.
	#@param $verifyPeer: if TRUE, will try to verify CA of CAS server
	#@param $verifyHostname: the verifyHostname level.  0 nonverification,
	#	1 just verify a CN in cert, 2 verify CN matches CAS server
	#	verified against certificate if $verifyPeer is set
	#@return TRUE on success, FALSE otherwise (in this later case, $err_msg
	#	contains an error message).
	#
	#This is heavily based on esup-phpCAS version, with added feature
	#of refactoring out the curl initialization to do more cert 
	#verification, etc.
	#
	{ 	$headers = '';
		$body = '';
		$err_msg = '';
		$res = TRUE;

		#initialize the cURL session
		$curl = $this->_curl_init($url, $cookies);

		# Now perform the query
		$buf = curl_exec ($curl);

		if ( $buf === FALSE ) 
		{ 	$err_msg = 'CURL error #' . curl_errno($curl) . 
				': '.curl_error($curl);
			#close the CURL session
			curl_close ($ch);
			$res = FALSE;
		} else 
		{ 	#close the CURL session
			curl_close ($curl);
	
			# find the end of the headers
			# note: strpos($str,"\n\r\n\r") does not work (?)
			$pos = FALSE;
			for ($i=0; $i<strlen($buf); $i++) 
			{   if ( $buf[$i] == chr(13) ) 
			    {	if ( $buf[$i+1] == chr(10) ) 
				{  if ( $buf[$i+2] == chr(13) ) 
				   { 	if ( $buf[$i+3] == chr(10) ) 
					{  	#header found
						$pos = $i;
						break;
					}
				   }
				}
			    }
			}
		}
	
		if ( $pos === FALSE ) 
		{ 	# end of header not found
			$err_msg = 'no header found';
			$res = FALSE;
		} else 
		{  	#extract headers into an array
			$headers = preg_split("/[\n\r]+/",substr($buf,0,$pos));	
			# extract body into a string
			$body = substr($buf,$pos+4);
		}

		return $res;
	}


	function _httpRedirect($newurl)
	#Redirects to $newurl
	#New URL is assumed to have already been detainted
	{	#Do a 303 redirect
		header("Location: $newurl",TRUE,303);
		#Just in case the above did not work
		?>
<html>
<head>	<title>Redirect</title>
	<script language="Javascript" type="text/javascript">
		window.location.href="<? $newurl ?>";
	</script>
	<meta http-equiv=refresh content="0; URL=<? $newurl ?>">
</head>
<body>
<h1>Redirect Requested</h1>
<p>This application is trying to redirect your browser, but apparently has
failed.  Please follow this link
<a href="<? $newurl ?>">to <? $newurl ?></a> manually.  Thank you.
</body>
</html>
<?php
		exit;
	}
		
#--------------------------------------------------------------------
#		Methods for dealing with URLs
#--------------------------------------------------------------------

	function _getServerBaseURL()
	#Returns the base URL for the CAS server
	{	$casURL = "https://" . $this->casServer .
			":" . $this->casPort;
		if ($this->casPath )
		{	$casURL = $casURL . '/' . $this->casPath;
		}
		return $casURL;
	}

	function _getServerLoginURL()
	#Returns the login URL for the CAS server
	{	$casURL = $this->_getServerBaseURL();
		$casURL="$casURL/login";
		return $casURL;
	}

	function _getServerLogoutURL()
	#Returns the login URL for the CAS server
	{	$casURL = $this->_getServerBaseURL();
		$casURL="$casURL/logout";
		return $casURL;
	}

	function _getServerServiceValidateURL()
	#Returns the login URL for the CAS server
	{	$casURL = $this->_getServerBaseURL();
		if ( $this->casVersion == SCAS_CAS_1_0 )
		{	$casURL = "$casURL/validate";
		} elseif ( $this->casVersion == SCAS_CAS_2_0 )
		{	$casURL = "$casURL/serviceValidate";
		} else
		{	fatal_error("Unknown CAS Version: " 
				. $this->casVersion);
			die("Should not reach here");
		}
		return $casURL;
	}

	function myURL()
	#Returns URL for current page.  Basically PHP_SELF after being
	#passed through html_entities.  This will likely fail if a hacker
	#tries some funny stuff, but XSS will be prevented.
	{	$path = $_SERVER['PHP_SELF'];
		$server = $_SERVER['SERVER_NAME'];
		$http = "http:";
		if ( IsSet($_SERVER['HTTPS']) && $_SERVER['HTTPS'] )
		{	$http = "https:";
		}
		$url = "$http//$server$path";
		$url = htmlentities($url);
		return $url;
	}

#--------------------------------------------------------------------
#		Methods for CAS redirect, service ticket validation
#--------------------------------------------------------------------

	function _redirectToCas($mode, $newurl=null)
	#This method is basically stolen from phpCAS, with some added stuff
	#to support renew mode and logout with optional redirection back
	#Mode can be set to one of:
	#	'gateway': do gateway log in (log in only if can do based
	#		on TGT cookie, never ask for user interaction)
	#	'renew': never use TGT cookie, force user to enter username/
	#		password
	#	'logout': logout from CAS, deleting TGT. Optionally return
	#		to $newurl
	#	'normal': standard, neither gateway, renew, or logout mode
	#Mode is case insensitive
	#
	#$newurl is the URL the CAS server should redirect to in order to
	#return.  For gateway, renew, and normal modes, this will default
	#to our url.  For logout, it will default to our URL unless set
	#and false, in which case no redirection after logout.
	{	$mode = strtolower($mode);
		$this->debug("Starting redirectToCas mode=$mode",
			SCAS_DBG_REDIRECT1);
		if ( ! IsSet($newurl) ||
			( $mode != 'logout' && ! $newurl ) )
		{	#Default newurl
			$newurl = $this->myURL();
		}
		$this->debug("newurl=$newurl", SCAS_DBG_REDIRECT2);

		switch ( $mode )
		{  case 'normal':
			$casURL = $this->_getServerLoginURL();
			$args = array('service' => $newurl);
			break;
		   case 'gateway':
			$casURL = $this->_getServerLoginURL();
			$args = array('service' => $newurl,
				'gateway' => 'true');
			break;
		   case 'renew':
			$casURL = $this->_getServerLoginURL();
			$args = array('service' => $newurl,
				'renew' => 'true');
			break;
		   case 'logout':
			$casURL = $this->_getServerLogoutURL();
			$args = array('url' => $newurl);
			break;
		}

		$redirectTo = $this->_constructURLWithArgs($casURL,$args);
		$this->debug("About to Redirect to $redirectTo", 
			SCAS_DBD_REDIRECT1);
		
		$this->_httpRedirect($redirectTo);
		die("Should not reach here");
	}

		
	function _have_valid_service_ticket($forced=FALSE)
	#Check to see if we have a valid service ticket.  
	#If so, generate authinfo stuff and return true
	#Else return false
	#
	#Modelled after esup phpCAS's validateST
	{	$this->debug("Starting have valid_service_ticket, " .
			"forced=$forced", SCAS_DBG_TICKET1);

		#Make sure we have a ticket
		if ( ! IsSet($_GET['ticket']) ) 
		{	$this->debug("No ticket, returning", SCAS_DBG_TICKET1);
			return FALSE;
		}
		$ticket = $_GET['ticket'];
		$this->debug("ticket is $ticket", SCAS_DBG_TICKET2);
		if (  ! IsSet($ticket) || ! $ticket ) 
		{	$this->debug("Ticket is empty or false, returning",
				SCAS_DBG_TICKET1);
			return FALSE;
		}

		#Generate callback URL to CAS
		$CASurl = $this->_getServerServiceValidateURL();
		$arry = array(	'ticket' => $ticket, 
				'service' => $this->myURL());
		if ( $forced )
		{	$arry['renew'] ='true';
		}
		$url = $this->_constructURLWithArgs($CASurl, $arry);

		#Read the URL
		$this->debug("About to ask CAS about ticket using url=$url", 
			SCAS_DBG_TICKET2);
		if ( ! $this->_readURL($url, '', #cookies
				$headers, $text_response, $err_msg,
				$this->casVerifyPeer, $this->casVerifyHost) )
		{	$this->_clearAuthInfo();
			$this->error("readURL returned error: $err_msg");
			return FALSE;
		}

		#Parse the returned information, depending on CAS version
		if ( $this->casVersion == SCAS_CAS_1_0 )
		{	if ( preg_match('/^no\n/',$text_response) )
			{	$this->_clearAuthInfo();
				$this->debug("CAS10 invalid ticket, returning",
					SCAS_DBG_TICKET1);
				return FALSE;
			}
			if ( ! preg_match('/^yes\n/', $text_response) )
			{	$this->_clearAuthInfo();
				$this->error("invalid CAS 1 response, " .
					$text_response . "\ntreating as no"
					);
				return FALSE;
			}
			#We have been validated
			$elements = preg_split('/\n/', $text_response);
			$username = trim($elements[1]);
			$this->_generateAuthInfo($username,$forced);
			$this->debug("returning with CAS10 valid ticket, " .
				"user=$username", SCAS_DBG_TICKET1);
			return TRUE;
		}

		if ( $this->casVersion != SCAS_CAS_2_0 )
		{	$this->fatal_error("Unknown CAS version " . 
				$this->casVersion);
			die("Should not reach here");
		}

		#CAS version 2.0, XML format, use DOM
		if ( ! ( $dom = domxml_open_mem($text_response) ) )
		{	$this->error("domxml_open_mem failed for response " .
				$text_response);
			$this->_clearAuthInfo();
			return FALSE;
		}

		#Read root note of XML tree and make sure is serviceResponse
		if ( ! ( $tree_response = $dom->document_element() ) )
		{	$this->error("DOM document_element failed");
			$this->_clearAuthInfo();
			return FALSE;
		}
		if ( $tree_response->node_name() != 'serviceResponse' )
		{	$this->error("Bad root node in CAS2 response" .
				$tree_response->node_name() );
			$this->_clearAuthInfo();
			return FALSE;
		}

		#Check for any authentication Failure sections
		#(order of checking failure vs success stanza reversed from
		#phpCAS)
		$elements = $tree_response->get_elements_by_tagname(
			'authenticationFailure');
		if ( sizeof($elements) != 0 )
		{	#Get the error code and message
			$errcode = $elements[0]->get_attribute('code');
			$errmsg = $elements[0]->get_content();
			$this->_clearAuthInfo();
			$this->debug("returning with CAS2 invalid ticket",
				SCAS_DBG_TICKET1);
			return FALSE;
		}

		#Now check for authSuccess sections
		$elements = $tree_response->get_elements_by_tagname(
			'authenticationSuccess');
		if ( sizeof($elements) > 1 )
		{	$this->error("Invalid XML in CAS2 response: multiple "
			. sizeof($elements) . "authSuccess tags");
			$this->_clearAuthInfo();
			return FALSE;
		}
		if ( sizeof($elements) == 1 )
		{	#Success, get username
			$elements = $elements[0]->get_elements_by_tagname(
				'user');
			if ( sizeof($elements) > 1 )
			{	$this->error("Invalid CAS2 authSuccess block: "
					. sizeof($elements) . "user tags");
				$this->_clearAuthInfo();
				return FALSE;
			}
			if ( sizeof($elements) < 1 )
			{	$this->error("Invalid CAS2 authSucc block: "
					. "no user tags");
				$this->_clearAuthInfo();
				return FALSE;
			}
			#Really a success, return TRUE
			$username = trim($elements[0]->get_content() );
			$this->_generateAuthInfo($username,$forced);
			$this->debug("returning with CAS2 valid ticket " .
				"for user=$username forced=$forced",
				SCAS_DBG_TICKET1);
			return TRUE;
		}

		#Should not reach here, neither authSuccess or autFailure
		#sections, invalid XML, so treat as failure
		$this->error("Invalid CAS2 response: no authSucc or authFail");
		$this->_clearAuthInfo();
		return FALSE;
	}

#--------------------------------------------------------------------
#		Upper level authentication, logout, etc. routines
#--------------------------------------------------------------------

	function authenticate()
	#Authenticates user.  Basically calls one of
	#	authenticate_optional
	#	authenticate_normal
	# or	authenticate_forced
	#depending on options in object
	#
	#This may or may not return.  Authenticate_normal and 
	#authenticate_forced should only return if authenticated, and
	#return the name of the authenticated user.
	#Authenticate_optional may return without authentication, in which
	#case returns FALSE, otherwise name of authenticated user.
	{	$this->debug("called authenticate", SCAS_DBG_AUTHENTICATE1);
		if ( $this->forcePassword)
		{	return $this->authenticate_forced();
		}
		if ( $this->authenticationOptional )
		{	return $this->authenticate_optional();
		}
		return $this->authenticate_normal();
	}

	function authenticate_forced()
	#Does a CAS renew type authentication.
	#If we have an authenticated session with the forced flag set, just
	#return user from that session.
	#If we have a 'ticket', try to authenticate it (with renew parm set)
	#and if valid, construct authinfo and return username.
	#Otherwise, redirect to CAS login in renew mode
	{	$this->debug("Called authenticate_forced", 
			SCAS_DBG_AUTHENTICATE1);

		if ( $this->is_authinfo_valid(TRUE) )
		{	$this->debug("Authenticated from session info",
				SCAS_DBG_AUTHENTICATE2);
			$this->_updateAuthInfoLastUsed(TRUE);
			return $this->username();
		}
		if ( $this->_have_valid_service_ticket(TRUE) )
		{	$this->debug("Authenticated from ST",
				SCAS_DBG_AUTHENTICATE2);
			return $this->username();
		}
		#No valid authinfo in session, no valid ST
		$this->debug("Redirecting to CAS in renew mode",
				SCAS_DBG_AUTHENTICATE2);
		$this->_redirectToCas('renew');
		die("Should not reach here");
	}

	function authenticate_normal()
	#Does a normal (non-renew, non-gateway) CAS authentication.
	#If we have an authenticated session, just
	#return user from that session.
	#If we have a 'ticket', try to authenticate it
	#and if valid, construct authinfo and return username.
	#Otherwise, redirect to CAS login
	{	$this->debug("Called authenticate_normal", 
			SCAS_DBG_AUTHENTICATE1);

		if ( $this->is_authinfo_valid() )
		{	$this->debug("Authenticated from session info",
				SCAS_DBG_AUTHENTICATE2);
			$this->_updateAuthInfoLastUsed(FALSE);
			return $this->username();
		}
		if ( $this->_have_valid_service_ticket() )
		{	$this->debug("Authenticated from ST",
				SCAS_DBG_AUTHENTICATE2);
			return $this->username();
		}
		#No valid authinfo in session, no valid ST
		$this->debug("Redirecting to CAS (normal)",
			SCAS_DBG_AUTHENTICATE2);
		$this->_redirectToCas('normal');
		die("Should not reach here");
	}

	function authenticate_optional()
	#Does a CAS gateway type authentication.
	#If we have an authenticated session, just #return user from that 
	#session.
	#If we have a 'ticket', try to authenticate it
	#and if valid, construct authinfo and return username.
	#Otherwise, redirect to CAS login in gateway mode
	#IF the gateway login succeeds, return username.
	#otherwise returns false.
	{	$this->debug("Called authenticate_optional", 
			SCAS_DBG_AUTHENTICATE1);

		if ( $this->is_authinfo_valid() )
		{	$this->debug("Authenticated from session info",
				SCAS_DBG_AUTHENTICATE2);
			$this->_updateAuthInfoLastUsed(FALSE);
			return $this->username();
		}
		if ( $this->_have_valid_service_ticket() )
		{	$this->debug("Authenticated from ST",
				SCAS_DBG_AUTHENTICATE2);
			return $this->username();
		}
		#No valid authinfo in session, no valid ST

		#Check for a timestamp from last visit to 
		#authenticate_optional.
		if ( IsSet($_SESSION[$this->sessionVarNameOptTstamp]) )
		{	$tstamp = $_SESSION[$this->sessionVarNameOptTstamp];
		} else
		{	$tstamp = 0;
		}
		if ( ! is_int($tstamp) ) $tstamp = 0;
		$delta = $this->authOptDeltaTime;
		$now = time();
		if ( $now < $tstamp + $delta )
		{	#Visited CAS server in gateway mode recently,
			#so don't revisit.  Since we are not authenticated,
			#just return
			$this->debug("Talked to CAS recently, return unauth",
				SCAS_DBG_AUTHENTICATE2);
			$this->_clearAuthInfo();
			return FALSE;
		}

		$_SESSION[$this->sessionVarNameOptTstamp] = time();
		$tstamp = $_SESSION[$this->sessionVarNameOptTstamp];
		$this->debug("Redirecting to CAS in gateway mode",
			SCAS_DBG_AUTHENTICATE2);
		$this->_redirectToCas('gateway');
		die("Should not reach here");
	}


	#Note that due to the (possibly silent) single sign-on mechanism of
	#CAS, logging out might not do what you expect on first thought

	function logoutSession()
	#Deletes the authinfo stuff for this session, requiring to visit
	#CAS server (although may be quiet if have a TGT cookie) before
	#reauthing
	{	$this->debug("Called logoutSession", SCAS_DBG_AUTHENTICATE1);
		$this->_clearAuthInfo();
		$_SESSION[$this->sessionVarNameOptTstamp] = 0;
	}

	function logoutCAS($newurl=null)
	#Does a CAS logout, causing users TGT cookie to be deleted. 
	#This generally should not be done and may be considered impolite 
	#(might affect other apps in a single-signon environment)
	#
	#If $newurl is given and not FALSE, will be sent to CAS server so
	#CAS server can refer user to this after logout.  If $newurl is
	#not set, will default to our URL.  Explicity give the value FALSE
	#to prevent CAS server from referring user to an URL
	#
	#NOTE: this function does not do logoutSession, so may wish to call
	#that first
	{	$this->debug("Called logoutCAS", SCAS_DBG_AUTHENTICATE1);
		if ( IsSet($newurl) )
		{	$this->debug("newurl is set to $newurl",
				SCAS_DBG_AUTHENTICATE2);
		} else
		{	$this->debug("newurl is not set, " .
				"will default to our URL", 
				SCAS_DBG_AUTHENTICATE2);
		} 

		$this->_redirectToCas('logout',$newurl);
	}

	function logoutForcedPrivs()
	#Causes the forced privs to be logged out/relinquished
	{	$this->debug("logoutForcedPrivs called", 
			SCAS_DBG_AUTHENTICATE1);
		$this->_unsetAuthInfoForced();
	}

	function logout()
	#Logout.  Will delete our session authinfo, 
	#If destroySessionOnLogout is true, will destroy the session
	#If casLogoutOnLogout is set and true, will do a CAS logout as well.
	{	$this->debug("Called logout", SCAS_DBG_AUTHENTICATE1);
		$this->logoutSession();
		if ( $this->destroySessionOnLogout )
		{	session_destroy();
		}
		if ( $this->casLogoutOnLogout)
		{	$this->logoutCAS();
		}
	}

	function username()
	#Returns the username of the authorized user.
	#Returns false if not authorized
	{	if ( ! IsSet($_SESSION[$this->sessionVarName]) )
			return FALSE;
		$sessionHash = $_SESSION[$this->sessionVarName];
		if ( ! IsSet($sessionHash['username']) )
			return FALSE;
		$username = $sessionHash['username'];
		return $username;
	}

}

?>
